package org.johnchamberlain.chess;

// a game represents one or more positions and their supplemental information
// game fragments such as problem positions and game excerpts are represented
// with instances of the game class as well as full games

// if the move list comprises a complete game then only the move list is persisted
// but if isolated positions cannot be constructed from the move list then those
//   positions will be stored as well
// when a game is loaded the position list is generated from the move list

public class Game {

	public static final int Version = 1;

    public static final int RESULT_Unknown = 0;
    public static final int RESULT_InProgress = 1;
    public static final int RESULT_NoResult = 2;
    public static final int RESULT_WhiteWon = 3;
    public static final int RESULT_BlackWon = 4;
    public static final int RESULT_Draw = 5;
    public static final int RESULT_DoubleForfeit = 6;
    public static final int RESULT_Incomplete = 7;
    private static final int RESULT_MIN = 0;
    private static final int RESULT_MAX = 7;

    public static final int RESULT_METHOD_Unknown = 0;
    public static final int RESULT_METHOD_Mate = 1;
    public static final int RESULT_METHOD_Resignation = 2;
    public static final int RESULT_METHOD_ForfeitByTime = 3;
    public static final int RESULT_METHOD_ForfeitByReferee = 4;
    public static final int RESULT_METHOD_DrawnAgreed = 5;
    public static final int RESULT_METHOD_DrawnRepetition = 6;
    public static final int RESULT_METHOD_DrawnByTime = 7;
    public static final int RESULT_METHOD_Drawn50Moves = 8;
    public static final int RESULT_METHOD_DrawnInsufficientMaterial = 9;
    public static final int RESULT_METHOD_DrawnInsufficientLosingChances = 10;
    public static final int RESULT_METHOD_DrawnByReferee = 11;
    public static final int RESULT_METHOD_DrawnOther = 12;
    public static final int RESULT_METHOD_Adjudication = 13;
    public static final int RESULT_METHOD_Abandonment = 14;
    public static final int RESULT_METHOD_Emergency = 15;
    public static final int RESULT_METHOD_DeathOrIncapacity = 16;
    private static final int RESULT_METHOD_MIN = 0;
    private static final int RESULT_METHOD_MAX = 15;

	private static final int NOTATION_Algebraic = 0;
	private static final int NOTATION_English = 0;
	// todo file/rank style, check style, ep style
	// todo language

	static final int MAX_MOVE_CAPACITY = 1000; // memory protection

	// transient private final static Engine theEngine = new Engine(); // the chess engine
    transient private final int[] aiTestBoard = new int[136 +1];     // used to test whether a move is possible

    private int _iResult = RESULT_Incomplete;
    private int _iResultMethod = RESULT_METHOD_Unknown;
    private GameInformation _gameInformation = null;  // this is stuff like the players and the tournaments etc
	private GameAnnotation _gameAnnotation = null;   // notes and annotations to the game

	private int miMoveCapacity = 50;
	private int miMoveLast = 0; // the last played move of the game (resigning etc is not considered a move)
    private int[] maiMoveList = new int[miMoveCapacity +1]; // item 0 is the size of list - by move number 1 - white's first move, 2 - black's first move etc.
    private int[] maiMoveListIrregular = new int[0];
        // shifted piece
        // piece removal
        // piece addition
        // no move
        // unknown move number

	/** the position is always the one resulting from the given move number
	 *  the initial position of the game is position 0
	 *  the position after the first move of the game is position 1
	 *  the final position is position miMoveLast
	 */
    private Position[] maPosition = new Position[miMoveCapacity+1];

    // format of a move
	// SPECIAL MOVES (always start with a move on same square 00-00, 10-10 etc)
	// 0000:0000    not a move
	// 0001:0000    move unknown
	// 0002:0000    no move played
	// 1100:0001    invalid request (such as specifying an illegal or unknown move number)
    // REQUIRED BITS
    // the lower 8 bits (1 to 8) is the destination square (0 to 63):
	// the first nibble (bits 1 to 4) are the file (a to f) and the second is the rank
    // the next 8 bits (9 to 16) are the origin square (0 to 63)
    // bits 17 to 18 are the promotion piece
    //  00 - knight
    //  01 - bishop
    //  10 - rook
    //  11 - queen
    // SUPPLEMENTAL BITS
    // bit 19 is the capture bit
    // bits 20 to 23 are the piece type captured
    //  0000 - none             1111 - unknown
    //  0001 - white pawn       1001 - black pawn
    //  0010 - white knight     1010 - black knight
    //  0011 - white bishop     1011 - black bishop
    //  0100 - white rook       1100 - black rook
    //  0101 - white queen      1101 - black queen
    //  0110 - white king       1110 - black king
    // bit 24 is the check bit
    // bits 25 to 28 are the piece moving
    //  (see piece type lists above)
    // bits 29 to 30 are the origin qualifier:
    //  00 - no origin specified (Ne6)
    //  01 - file origin only (Nfe6)
    //  10 - rank origin only (N4e6)
    //  11 - full origin (Qe7e6)
    // bits 31 and 32 specify castling:
    //  00 - no castling
    //  01 - kingside castling (0-0)
    //  10 - queenside castling (0-0-0)
    //  11 - unknown

	public static final int MOVE_NotaMove       = 0x00000000;
	public static final int MOVE_MoveUnknown    = 0x00010000;
	public static final int MOVE_NoMovePlayed   = 0x00020000;
	public static final int MOVE_InvalidRequest = 0x00030000;

	static final int SQUARE_FileMask  = 0x0F; // there is no shift for the file
	static final int SQUARE_RankMask  = 0xF0;
	static final int SQUARE_RankShift = 4;

	public static final int MOVE_DestinationFileMask = 0x0000000F;
	public static final int MOVE_DestinationRankMask = 0x000000F0, MOVE_DestinationRankShift = 4;
	public static final int MOVE_OriginFileMask = 0x00000F00, MOVE_OriginFileShift = 8;
	public static final int MOVE_OriginRankMask = 0x0000F000, MOVE_OriginRankShift = 12;

	public static final int MOVE_DestinationMask      = 0x000000FF;
	public static final int MOVE_OriginMask           = 0x0000FF00, MOVE_OriginShift = 8;
	public static final int MOVE_CaptureMask          = 0x00040000, MOVE_CaptureShift = 18;
	public static final int MOVE_PieceCapturedMask    = 0x00780000, MOVE_PieceCapturedShift = 19;
	public static final int MOVE_CheckMask            = 0x00800000, MOVE_CheckShift = 23;
	public static final int MOVE_PieceMovingMask      = 0x0F000000, MOVE_PieceMovingShift = 24;
	public static final int MOVE_PieceNCMovingMask    = 0x07000000; // no color specified
	public static final int MOVE_CastlingMask         = 0xC0000000, MOVE_CastlingShift = 30;
	public static final int MOVE_CastlingKingSide     = 0xA0000000;
	public static final int MOVE_CastlingQueenSide    = 0xB0000000;
	public static final int MOVE_OriginQualifierMask  = 0x30000000;
	public static final int MOVE_OriginQualifier_File = 0x10000000;
	public static final int MOVE_OriginQualifier_Rank = 0x20000000;
	public static final int MOVE_OriginQualifier_Both = 0x30000000;
	public static final int MOVE_OriginQualifier_None = 0x00000000;

	public static final int MOVE_PromotionMask = 0x00030000, MOVE_PromotionShift = 16;
	public static final int PIECE_Promotion_Queen     = 0x0;
	public static final int PIECE_Promotion_Rook      = 0x1;
	public static final int PIECE_Promotion_Bishop    = 0x2;
	public static final int PIECE_Promotion_Knight    = 0x3;
	public static final int PIECE_Promotion_Queen_NS  = 0x00000; // no shift
	public static final int PIECE_Promotion_Rook_NS   = 0x10000; // no shift
	public static final int PIECE_Promotion_Bishop_NS = 0x20000; // no shift
	public static final int PIECE_Promotion_Knight_NS = 0x30000; // no shift

	public static final int PIECE_None    = 0x00000000;
	public static final int PIECE_Unknown = 0x000000;
	public static final int PIECE_Pawn_NoShift    = 0x1000000; // white
	public static final int PIECE_Pawn    = 0x1; // white
	public static final int PIECE_Knight  = 0x2; // white
	public static final int PIECE_Bishop  = 0x3; // white
	public static final int PIECE_Rook    = 0x4; // white
	public static final int PIECE_Queen   = 0x5; // white
	public static final int PIECE_King    = 0x6; // white
	public static final int PIECE_BPawn   = 0x9;
	public static final int PIECE_BKnight = 0xA;
	public static final int PIECE_BBishop = 0xB;
	public static final int PIECE_BRook   = 0xC;
	public static final int PIECE_BQueen  = 0xD;
	public static final int PIECE_BKing   = 0xE;

	public static final char[] aPIECE_Indicator = new char[7];
	static {
		aPIECE_Indicator[0] = '?';
		aPIECE_Indicator[1] = 'P';
		aPIECE_Indicator[2] = 'N';
		aPIECE_Indicator[3] = 'B';
		aPIECE_Indicator[4] = 'R';
		aPIECE_Indicator[5] = 'Q';
		aPIECE_Indicator[6] = 'K';
	}

    public Game() {}

	static String sRENDER_Result( int iResult ){
		switch(iResult) {
			case RESULT_Unknown: return "[unknown]";
			case RESULT_InProgress: return "[other]";
			case RESULT_NoResult: return "[no result]";
			case RESULT_WhiteWon: return "1-0";
			case RESULT_BlackWon: return "0-1";
			case RESULT_Draw: return "-";
			case RESULT_DoubleForfeit: return "0-0";
			case RESULT_Incomplete: return "[incomplete]";
			default: return "[?]";
		}
	}

	static String sRENDER_ResultMethod( int iResultMethod ){
		switch(iResultMethod) {
		    case RESULT_METHOD_Unknown: return "[unknown]";
		    case RESULT_METHOD_Mate: return "mate";
		    case RESULT_METHOD_Resignation: return "resigns";
		    case RESULT_METHOD_ForfeitByTime: return "forfeit";
		    case RESULT_METHOD_ForfeitByReferee: return "forfeit by referee";
		    case RESULT_METHOD_DrawnAgreed: return "draw agreed";
		    case RESULT_METHOD_DrawnRepetition: return "drawn by repetition";
		    case RESULT_METHOD_DrawnByTime: return "drawn by time";
		    case RESULT_METHOD_Drawn50Moves: return "draw [50 moves]";
		    case RESULT_METHOD_DrawnInsufficientMaterial: return "drawn [lack of mating material]";
		    case RESULT_METHOD_DrawnInsufficientLosingChances: return "drawn [insufficient losing chances]";
		    case RESULT_METHOD_DrawnByReferee: return "drawn by referee";
		    case RESULT_METHOD_DrawnOther: return "draw";
		    case RESULT_METHOD_Adjudication: return "drawn by adjudication";
		    case RESULT_METHOD_Abandonment: return "game abandoned";
		    case RESULT_METHOD_Emergency: return "game aborted";
		    case RESULT_METHOD_DeathOrIncapacity: return "incapacitated";
			default: return "[?]";
		}
	}

	void vClear(){
		Engine theEngine = new Engine();
		_iResult = RESULT_Incomplete;
		_iResultMethod = RESULT_METHOD_Unknown;
		_gameInformation = null;
		_gameAnnotation = null;
		miMoveLast = 0;
		maiMoveList = new int[miMoveCapacity +1];
		maiMoveListIrregular = new int[0];
		Position[] maPosition = new Position[miMoveCapacity+1];
	}

	boolean zSetToInitialPosition(StringBuffer sbError){
		this.vClear();
		this.maPosition[0] = positionGenerate(0, sbError); // todo raise error
		return (this.maPosition[0] == null) ? false : true;
	}

	/**
	 *  @param iMoveNumber  the move made to create the position; initial position is number 0
	 */
	Position getPosition(int iMoveNumber){
		if(iMoveNumber > miMoveLast) return null;
		if(iMoveNumber < 0) return null;
		return maPosition[iMoveNumber];
	}

	// todo irregular moves
	int getMove(int iMoveNumber){
		if( iMoveNumber < 1 ) return MOVE_InvalidRequest;
	    if( iMoveNumber > miMoveLast ) return MOVE_InvalidRequest;
		return maiMoveList[iMoveNumber];
	}

	String getMoveNotation( int xMove ){
		return getNotationForMoveCode(getMove(xMove));
	}

	// todo checkmate (#) notation
	static char[] sacMove = new char[12];
	static int slenMove = 0;
	static String getNotationForMoveCode( int iMove ){
		int iOriginFile, iDestinationFile, iDestinationRank;
		if( iMove == MOVE_NotaMove ) return "ERR";
		if( iMove == MOVE_MoveUnknown ) return "-";
		if( iMove == MOVE_NoMovePlayed ) return "[x]";
		if( iMove == MOVE_InvalidRequest ) return "ERR?";
		iOriginFile      = (iMove & MOVE_OriginFileMask) >> MOVE_OriginFileShift;
		iDestinationFile = (iMove & MOVE_DestinationFileMask);
		iDestinationRank = ((iMove & MOVE_DestinationRankMask) >> MOVE_DestinationRankShift);
		if( (iMove & MOVE_PieceNCMovingMask) == PIECE_Pawn_NoShift ){
			if( iOriginFile == iDestinationFile ){
				sacMove[0] = (char)(iOriginFile + 0x60);
				sacMove[1] = (char)(iDestinationRank + 0x30);
				slenMove = 2;
			} else { // capture
				sacMove[0] = (char)(iOriginFile + 0x60);
				sacMove[1] = (char)(iDestinationFile + 0x60);
				slenMove = 2;
				int iOriginQualifierNoShift = iMove & MOVE_OriginQualifierMask; // if pawns are doubled captures can be ambiguous
				if( iOriginQualifierNoShift > 0 ){
					sacMove[++slenMove-1] = (char)(iDestinationRank + 0x30);
				}
			}
			if( iDestinationRank == 8 || iDestinationRank == 1 ){ // promotion
				sacMove[++slenMove-1] = '=';
				int iPromotionPiece = (iMove & MOVE_PromotionMask); // no shift
				if( iPromotionPiece == PIECE_Promotion_Queen_NS )
					sacMove[++slenMove-1] = 'Q'; // todo conflate with array above
				else if ( iPromotionPiece == PIECE_Promotion_Rook_NS )
					sacMove[++slenMove-1] = 'R'; // todo conflate with array above
				else if ( iPromotionPiece == PIECE_Promotion_Bishop_NS )
					sacMove[++slenMove-1] = 'B'; // todo conflate with array above
				else if ( iPromotionPiece == PIECE_Promotion_Knight_NS )
					sacMove[++slenMove-1] = 'N'; // todo conflate with array above
				else
					sacMove[++slenMove-1] = '?'; // todo conflate with array above
			}
			return new String(sacMove, 0, slenMove);
		} else {
			int iPiece = (iMove & MOVE_PieceNCMovingMask) >> MOVE_PieceMovingShift;
			if( iPiece == PIECE_King ) // check for castling
				if( iOriginFile == 5){
					if( iDestinationFile == 7 ) return "0-0";
					if( iDestinationFile == 3 ) return "0-0-0";
				}
			char cPieceIndicator = aPIECE_Indicator[iPiece];
			int iOriginQualifierNoShift = iMove & MOVE_OriginQualifierMask;
		    if( iOriginQualifierNoShift == 0 )
				return new String(new StringBuffer().append(cPieceIndicator).append((char)(iDestinationFile + 0x60)).append((char)(iDestinationRank + 0x30)));
		    else
				if( iOriginQualifierNoShift == MOVE_OriginQualifier_File )
					return new String(new StringBuffer().append(cPieceIndicator).append((char)(iOriginFile + 0x60)).append((char)(iDestinationFile + 0x60)).append((char)(iDestinationRank + 0x30)));
				else {
					int iOriginRank = ((iMove & MOVE_OriginRankMask) >> MOVE_OriginRankShift);
					return new String(new StringBuffer().append(cPieceIndicator).append((char)(iOriginRank + 0x30)).append((char)(iDestinationFile + 0x60)).append((char)(iDestinationRank + 0x30)));
				}
		}
	}

	int getLastMoveNumber(){ return miMoveLast; }

    public java.util.Iterator getMoveIterator(){
    	return null;
    }

    public int getResult(){return _iResult;}
    public void setResult(int iResult){
    	if(iResult<RESULT_MIN || iResult>RESULT_MAX) return;
    	_iResult = iResult;
     	return;
    }

    public int getResultMethod(){return _iResultMethod;}
    public void setResultMethod(int iResultMethod){
    	if(iResultMethod<RESULT_METHOD_MIN || iResultMethod>RESULT_METHOD_MAX) return;
    	_iResultMethod = iResultMethod;
     	return;
	}

	public String getSquareName(int iSquare){
		int iRank = (iSquare & 0xF0) >> 4;
		int iFile = (iSquare & 0x0F);
		char cFile = (char)(0x60 + iFile);
		return new String("" + cFile + iRank);
	}

    public GameInformation getGameInformation(){return _gameInformation;}

    public GameAnnotation getGameAnnotation(){return _gameAnnotation;}

    /** Makes a move in this game.
     *  Move number 1 is White's first move, move number 2 is Black's first move and so on.
     */
    public boolean zWriteMove(int iMoveNumber, int sqOrigin, int sqDest, int piecePromotion, StringBuffer sbError){
    	try {
			if(iMoveNumber < 1){
				sbError.append("illegal move number [" + iMoveNumber + "]; move number must be greater than 0");
				return false;
			}
			if(sqOrigin < 1 || sqOrigin > 0x88){
				sbError.append("illegal origin square [" + sqOrigin + "]; must be between 1-64");
				return false;
			}
			if(sqDest < 1 || sqDest > 0x88){
				sbError.append("illegal destination square [" + sqDest + "]; must be between 1-64");
				return false;
			}
			if(piecePromotion < PIECE_Promotion_Queen || piecePromotion > PIECE_Promotion_Knight){
				sbError.append("illegal piece promotion [" + piecePromotion + "]");
				return false;
			}
			if(iMoveNumber > miMoveCapacity){
				if(zIncreaseMoveCapacity(iMoveNumber, sbError)){
					// continue
				} else {
					sbError.insert(0, "unable to increase the move capacity to " + iMoveNumber + ": ");
					return false;
				}
			}
			int iMove = moveFormulate(maPosition[iMoveNumber-1], sqOrigin, sqDest, piecePromotion, sbError);
		    if( iMove == 0 ){
				sbError.insert(0, "Failed to formulate move [" + sqOrigin + "-" + sqDest + ":" + piecePromotion + "]: ");
				return false;
			}
    		maiMoveList[iMoveNumber] = iMove;
			if( iMoveNumber > miMoveLast ) miMoveLast = iMoveNumber;
		    Position newPosition = positionGenerate(iMoveNumber, sbError);
			if(newPosition == null){ // todo in future may permit entry of blind moves
				sbError.insert(0, "unable to generate position: ");
				return false;
			} else {
				maPosition[iMoveNumber] = newPosition;
		    }
			return true;
		} catch(Exception ex) {
			Utility.errorUnexpected(ex, sbError);
			return false;
		}
    }

	String sMoveDetail(int iMove){
		StringBuffer sbMove = new StringBuffer(36);
		sbMove.append( this.getSquareName((iMove & MOVE_OriginMask) >> MOVE_OriginShift) );
		sbMove.append( '-' );
		sbMove.append( this.getSquareName(iMove & MOVE_DestinationMask) );
		sbMove.append( ':' );
		sbMove.append( this.sMovePromotionPiece((iMove & MOVE_PromotionMask) >> MOVE_PromotionShift) );
		sbMove.append( ':' );
		sbMove.append( this.sMovePiece((iMove & MOVE_PieceMovingMask) >> MOVE_PieceMovingShift) );
		sbMove.append( ':' );
		int iCaptureBit = iMove & MOVE_CaptureMask;
		if( iCaptureBit > 0 ){
			sbMove.append("x" + this.sMovePiece((iMove & MOVE_PieceCapturedMask) >> MOVE_PieceCapturedShift) );
		} else {
			sbMove.append("..");
		}
		sbMove.append( ':' );
		sbMove.append(this.sMoveOriginQualifier(iMove & MOVE_OriginQualifierMask));
		return sbMove.toString();
	}

	String sMovePromotionPiece(int iPromotionPiece){
		switch(iPromotionPiece){
			case PIECE_Promotion_Queen  : return "Q";
			case PIECE_Promotion_Rook   : return "R";
			case PIECE_Promotion_Bishop : return "B";
			case PIECE_Promotion_Knight : return "N";
			default                     : return "?";
		}
	}

	String sMovePiece(int iPiece){
		switch(iPiece){
			case PIECE_King    : return "WK";
			case PIECE_Queen   : return "WQ";
			case PIECE_Rook    : return "WR";
			case PIECE_Bishop  : return "WB";
			case PIECE_Knight  : return "WN";
			case PIECE_Pawn    : return "WP";
			case PIECE_BKing   : return "BK";
			case PIECE_BQueen  : return "BQ";
			case PIECE_BRook   : return "BR";
			case PIECE_BBishop : return "BB";
			case PIECE_BKnight : return "BN";
			case PIECE_BPawn   : return "BP";
			default                     : return "??";
		}
	}

	String sMoveOriginQualifier(int iOriginQualifier){
		switch(iOriginQualifier){
			case MOVE_OriginQualifier_File : return "F-";
			case MOVE_OriginQualifier_Rank : return "-R";
			case MOVE_OriginQualifier_Both : return "FR";
			case MOVE_OriginQualifier_None : return "--";
			default                        : return "??";
		}
	}

	/** This function adds the supplemental information bits to the move.
	 *  pass 0:0 to indicate that no move was played
	 *  Returns 0 on an error */
	int moveFormulate(Position thePosition, int sqOrigin, int sqDest, int piecePromotion, StringBuffer sbError){
		if( thePosition == null ){ sbError.append("the position is null"); return 0; }
		if( sqOrigin==0 && sqDest==0) return MOVE_NoMovePlayed;
		int iMove = 0;

		// record move
		int iOriginFile = sqOrigin & 0x0F;
		int iOriginRank = (sqOrigin & 0xF0) >> 4;
		int iDestinationFile = sqDest & 0x0F;
		int iDestinationRank = (sqDest & 0xF0) >> 4;
		if( iOriginFile<1 || iOriginFile>8 || iOriginRank<1 || iOriginRank>8 ){ sbError.append("invalid origin [0x" + Integer.toHexString(sqOrigin) + "]"); return 0; }
		if( iDestinationFile<1 || iDestinationFile>8 || iDestinationRank<1 || iDestinationRank>8 ){ sbError.append("invalid destination [0x" + Integer.toHexString(sqDest) + "]"); return 0; }
		iMove |= sqOrigin << MOVE_OriginShift;
		iMove |= sqDest; // no shift required

		// verify that there is an origin piece
		int pieceOrigin = thePosition.aiBoard[sqOrigin];
		if( pieceOrigin == PIECE_None ){ sbError.append("origin square [0x" + Integer.toHexString(sqOrigin) + "] contains no piece"); return 0; }
		int pieceType = pieceOrigin & 7;

		// set the promotion bit if there was a promotion
		if( piecePromotion > 0 ) iMove |= (piecePromotion << MOVE_PromotionShift);

		// set the capture bit and piece captured if there was a capture
		if( thePosition.aiBoard[sqDest] == PIECE_None ){ // check for an en passant capture
			if( (pieceType == PIECE_Pawn) && (iOriginFile != iDestinationFile) ){
				iMove |= MOVE_CaptureMask; // set the capture bit
				int iPieceCaptured = thePosition.zWhiteToMove ? PIECE_BPawn : PIECE_Pawn;
				iMove |= (iPieceCaptured << MOVE_PieceCapturedShift); // set the piece being captured
			}
		} else {
			iMove |= MOVE_CaptureMask; // set the capture bit
			int iPieceCaptured = thePosition.aiBoard[sqDest];
			iMove |= (iPieceCaptured << MOVE_PieceCapturedShift); // set the piece being captured
		}

		// set the check bit if the opponent is in check at the end of the move
		theEngine.copyBoard(thePosition.aiBoard, aiTestBoard);
		boolean zOpponentInCheck = false;
		if(thePosition.zWhiteToMove){
			zOpponentInCheck = theEngine.zInCheck_Black(aiTestBoard);
		} else {
			zOpponentInCheck = theEngine.zInCheck_White(aiTestBoard);
		}
		if( zOpponentInCheck ) (iMove |= MOVE_CheckMask);

		// set the piece moving
		iMove |= (pieceOrigin << MOVE_PieceMovingShift);

		// specify castling
		if( (pieceType == PIECE_King) && (iOriginFile == 5) && (iDestinationFile == 3 || iDestinationFile == 7) ){ // castling
			if( iDestinationFile == 7 ){
				iMove |= MOVE_CastlingKingSide;
			} else {
				iMove |= MOVE_CastlingQueenSide;
			}
		} else { // not castling - specify origin qualifier if necessary // todo logic is different for English notation

			// (1) check to see if there is another piece of the same kind (homolog), for each
			// (1a) can it move to the dest square?
			// (1b) will the mover be in check if the move is made?
			// (2) is the moving piece on a unique file compared to the other candidates? if yes, use it
			// (3) if no, is the moving piece on a unique rank compared to others? if yes, use it
			// (4) if no, then specify both file and rank

		    // identify homologs (note that a smarter version of this could use directionals
			// from the destination square to identify homologs)
			boolean zHomologFile = false, zHomologRank = false;
			if( pieceType == PIECE_Pawn ){
				for(int xSquare = 1; xSquare <= 0x88; xSquare++){
					if( xSquare == sqOrigin ) continue; // do not include the moving piece obviously
					if( thePosition.aiBoard[xSquare] == pieceOrigin ){
						if(( iOriginFile != (xSquare & SQUARE_FileMask)) ) continue; // pawns on different files
						int sqPawnDestination;
						if( thePosition.zWhiteToMove ){
							if( iDestinationFile < iOriginFile ) sqPawnDestination = xSquare+0x0F;
							else                                 sqPawnDestination = xSquare+0x11;
						} else {
							if( iDestinationFile < iOriginFile ) sqPawnDestination = xSquare-0x11;
							else                                 sqPawnDestination = xSquare+0x0F;
						}
						if(theEngine.zIsLegalMove(thePosition, xSquare, sqPawnDestination)){ // found a viable homolog
							zHomologFile = true;
							break; // a pawn can only have a file homolog
						}
					}
				}
			} else { // moving a piece
				for(int xSquare = 1; xSquare <= 0x88; xSquare++){
					if( xSquare == sqOrigin ) continue; // do not include the moving piece obviously
					if( thePosition.aiBoard[xSquare] == pieceOrigin ){
						if(theEngine.zIsLegalMove(thePosition, xSquare, sqDest)){ // found a viable homolog
							if( iOriginFile == (xSquare & SQUARE_FileMask)) zHomologFile = true;
							if( iOriginRank == ((xSquare & SQUARE_RankMask) >> SQUARE_RankShift)) zHomologRank = true;
						}
					}
				}
			}
			if( zHomologFile || zHomologRank )
				if( zHomologFile )
					if( zHomologRank )
						iMove |= MOVE_OriginQualifier_Both; // set both
					else
						iMove |= MOVE_OriginQualifier_Rank;
				else
					iMove |= MOVE_OriginQualifier_File;
		}

		return iMove;
	}

	// exceptions bubble up
	private Position positionGenerate(int iMoveNumber, StringBuffer sbError){
		if( iMoveNumber == 0 ){
	        this.theEngine.setPositionInitial();
	    	return this.theEngine.getPosition();
		}
		if( iMoveNumber > miMoveCapacity ) return null;
		Position positionPrevious = maPosition[iMoveNumber - 1];
		if( positionPrevious == null ){ sbError.append("no previous position"); return null; }// currently must have previous position to generate (todo add reconstruct feature)
		int iMove = maiMoveList[iMoveNumber];
		if( iMove == MOVE_NotaMove ){ sbError.append("not a move"); return null; }
		if( iMove == MOVE_MoveUnknown ){ sbError.append("move unknown"); return null; }
		if( iMove == MOVE_InvalidRequest ){ sbError.append("invalid request"); return null; }
		if( iMove == MOVE_NoMovePlayed ) return positionPrevious.rclone();
		Position positionNew = positionPrevious.rclone();

		if(!theEngine.zsetPosition(positionNew, sbError)){ sbError.insert(0, "unable to set position: "); return null; }

        int sqFrom = (iMove & MOVE_OriginMask) >> MOVE_OriginShift;
        int sqTo   = iMove & MOVE_DestinationMask;
        int piecePromotion = (iMove & MOVE_PromotionMask) >> MOVE_PromotionShift;
		if(!theEngine.zTransform_MakeMove(sqFrom, sqTo, piecePromotion, sbError)){ sbError.insert(0, "unable to make move: "); return null; }

		return positionNew;

	}

	private boolean zIncreaseMoveCapacity(int iNewCapacityMinimum, StringBuffer sbError){
		sbError.append("not supported");
		return false;
	}

}